//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: Client
//     Author: Jim Guggemos
//     Description: Class that handles client socket communication.
//
// $Date$
//
//***************************************************************************

#include "Client.h"
#include "Socket.h"
#include "SAPQ.h"
#include <stdarg.h>
#include "HashManager.h"
#include <string.h>
#include <limits.h>
#include <sstream>

Client::Client(HashManager &hashmgr, Socket *clientSock) :
    m_socket(clientSock), m_threadId(0), m_hashmgr(hashmgr)
{
}

Client::~Client()
{
    if ( m_socket )
        delete m_socket;

    m_socket = 0;
}

bool
Client::start()
{
    pthread_attr_t attr;
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
    if ( pthread_create( &m_threadId, &attr, threadRun, (void*)this ) != 0 )
        return false;

    pthread_attr_destroy( &attr );		// No effect in LINUX

    return true;
}

void
Client::run()
{
    logDebug( "Client connection opened" );

    while ( 1 )
    {
        string command;

        if ( ! readSocket( command ) )
            break;

        if ( command.empty() )
            continue;

        vector<string> args;
        string arg;


        // Split command into args
        stringstream ss(command, ios_base::in);
        while(getline(ss, arg, ' '))
            args.push_back(arg);

        if( args.empty() )
            continue;

        try
        {
            if ( args[0]  == "clear" )
            {
                m_hashmgr.clear();
                ok();
                logDebug( "MAC hash table cleared" );
            }

            else if ( args[0] == "ping" )
                ok();

            else if ( args[0] == "allow" )
            {
                HwAddr hw( args[1] );
                if( args.size() > 2 )
                    hw.setTag(args[2].c_str());

                m_hashmgr.put( hw );
                ok();
                logDebug( "MAC '%s' added to hash table", hw.ascii().c_str() );
            }

            else if ( args[0] == "deny" )
            {
                if(args.size() > 1)
                {
                    HwAddr hw( args[1] );
                    m_hashmgr.remove( hw );
                    ok();
                    logDebug( "MAC '%s' removed from hash table", hw.ascii().c_str() );
                }
            }

            else if ( args[0] == "list" )
            {
                if(args.size() > 1)
                    list(args[1]);
                else
                    list("");
            }

            else if ( args[0] == "stats" )
            {
                if(args.size() > 1)
                    stats(args[1]);
                else
                  stats(string(""));
            }

            else if ( args[0] == "debug" )
            {
                if(args.size() > 1)
                    debug(args[1]);
                else
                    debug("");
                ok();
            }

            else if ( args[0] == "help" || args[0] == "?" )
            {
                help();
                ok();
            }

            else
                throw invalid_argument( "syntax error" );
        }
        catch (std::exception &e)
        {

            error( "error in command '%s': %s", command.c_str(), e.what() );
            logDebug( "error in command '%s': %s", command.c_str(), e.what() );
        }
    }

    logDebug( "Client connection closed" );
}

void
Client::ok()
{
    m_socket->send( "ok\n" );
}

void
Client::error(const char *format, ...)
{
    char buf[4096];

    va_list ap;
    va_start( ap, format );
    vsnprintf( buf, 4095, format, ap );
    buf[4094] = 0;
    strcat( buf, "\n" );
    va_end( ap );

    m_socket->send( buf );
}

void
Client::list(string arg)
{
    char *endptr;
    char buf[100];
    int size = -1;
    
    HwAddrHash::AddrList alist;

    
    if(!arg.empty())
    {
        size = strtol(arg.c_str(), &endptr, 10);
        if ((errno == ERANGE && (size == LONG_MAX || size == LONG_MIN)) || (errno != 0 && size == 0))
        {
            strerror_r(errno, buf, sizeof(buf));
            error("Invalid argument: %s", buf);
            return;
        }
        else if(size < 1 || size > 6)
        {
            error("Invalid argument: %d.  Please specify a number between 1 and 6.", size);
            return;
        }
    }
    
    if ( m_hashmgr.getAll( alist, size ) )
    {
        HwAddrHash::AddrList::const_iterator i = alist.begin();
        const HwAddrHash::AddrList::const_iterator iEnd = alist.end();

        for ( ; i != iEnd; ++i )
        {
            const HwAddr *addr = &(*i);
            string s(addr->ascii());
            
            s.append(18 - s.size(), ' ');
            s.append(addr->getTag());
            s.append("\n");
            m_socket->send(s);
        }
    }
    ok();
}

void
Client::stats(string arg)
{
    char *endptr;
    char buf[100];
    int size = -1;

    if(!arg.empty())
    {
        size = strtol(arg.c_str(), &endptr, 10);
        if ((errno == ERANGE && (size == LONG_MAX || size == LONG_MIN)) || (errno != 0 && size == 0))
        {
            strerror_r(errno, buf, sizeof(buf));
            error("Invalid argument: %s", buf);
            return;
        }
        else if(size < 1 || size > 6)
        {
            error("Invalid argument: %d.  Please specify a number between 1 and 6.", size);
            return;
        }
    }
    
    string stats = m_hashmgr.generateStats(size);
    m_socket->send( stats );
    ok();
}

void
Client::debug(string arg)
{
    if (!arg.empty())
    {
        if ( arg == "yes" || arg == "true" || arg == "on" )
            setDebugMode(true);

        else if ( arg == "no" || arg == "false" || arg == "off" )
            setDebugMode(false);

        else
            throw invalid_argument("invalid argument to debug: '"+arg+"'");
    }

    string result = "Debug mode is ";
    result += (debugMode() ? "on" : "off");
    m_socket->send( result+"\n" );
}

void
Client::help()
{
    string result;

    result += "Valid commands are:\n";
    result += "  clear - clear all entries for allowed MACs table\n";
    result += "  ping - replies with ok\n";
    result += "  allow <mac> [tag] - add <mac> to allowed MACs table.  If tag is specified, it will be written to the outgoing packet's BOOTP filename field\n";
    result += "  allow <prefix> [tag] - add <prefix> (XX, XX:XX, etc) to allowed MACs table.  If tag is specified, it will be written to the outgoing packet's BOOTP filename field\n";
    result += "  deny <mac> - remove <mac> from allowed MACs table\n";
    result += "  deny <prefix> - remove <prefix> (XX, XX:XX, etc) from allowed MACs table\n";
    result += "  list [size] - list all entries in allowed MACs table of size [size].  No arg to show all MACs and prefixes\n";
    result += "  stats [size] - return statistics for table of size [size].  No arg to show all table statistics\n";
    result += "  debug [on|off] - turn debug logging on/off (no arg to show current)\n";
    result += "  help - show this list\n";
    m_socket->send( result );
}

bool
Client::readSocket(string &command)
{
    const int max_length = 1024;
    bool result = false;

    command.clear();

    try
    {
	command = m_socket->recv_line( max_length );
	result = true;
    }
    catch ( Socket::LineTooLongException &e )
    {
        logMessage( "Client socket: line too long (> %d bytes)",
                    max_length );

	// Even though this is an error, it is not an exit condition.
	result = true;
    }
    catch ( Socket::IncompleteLastLineException &e )
    {
        logMessage( "Client socket: received incomplete last command upon "
                    "closure: '%s'", e.data().c_str() );
    }
    catch ( Socket::ClosedException )
    {
    }
    catch ( Socket::Exception &e )
    {
        logMessage( "Client socket: socket error: %s", e.getMsg().c_str() );
    }

    // trim white space from right end
    
    if ( ! command.empty() )
    {
        int idx = command.length() - 1;
        for ( ; idx >= 0 && isspace(command[idx]); --idx ) ;
        command.erase( idx+1 );
    }

    return result;
}


void *
Client::threadRun(void *vthis)
{
    Client *rthis = reinterpret_cast<Client *>(vthis);

    // Will delete the Client object when the scope exits, including when
    // thread is killed, which will close the socket.
    class chelper
    {
    private:
        Client *m_cli;
        
    public:
        chelper(Client *c) : m_cli(c) {}
        ~chelper() { delete m_cli; }
    } _cli(rthis);

    rthis->run();

    return NULL;
}
