//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: HwAddrHash
//     Author: Jim Guggemos
//     Description: Classes to maintain thread-safe hash table of allowable
//                  hardware addresses.
//
// $Date$
//
//***************************************************************************

#include "HwAddrHash.h"
#include "SAPQ.h"
#include <map>
#include <string.h>

using namespace __gnu_cxx;

#define LOCK() Mutex::Holder __lock(m_mutex)


// ---- HwAddr Class Definition --------------

HwAddr::HwAddr() throw()
{
    memset( (void*)bytes, 0, 6 );
    memset( tag, '\0', sizeof(tag) );
    addrsize = 6;
    // This is technically an "invalid" hw addr (checkValidity() would
    // throw an exception) but this is only used to create a HwAddr
    // before overwriting the data.
}


HwAddr::HwAddr(const unsigned char *buf, int len/*=6*/) throw(std::invalid_argument) : addrsize(len)
{
    memset( (void*)bytes, 0, 6 );
    memset( tag, '\0', sizeof(tag) );
    cpyMem(buf);
    checkValidity();
}

HwAddr::HwAddr(const string &ascii) throw(std::invalid_argument)
{
    memset( tag, '\0', sizeof(tag) );

    const char *p = ascii.c_str();
    const char *end = p + strlen(p);
    
    addrsize = 0;
    for ( int i = 0; i < 6; ++i )
    {
        bytes[i] = '\0';
        
        if(p < end - 1)
        {
            addrsize++;
            for ( int j = 0; j < 2; ++j )
            {
                unsigned char cval = hexval( *p++ );
                if ( cval > 15 )
                    throw invalid_argument( "ascii hardware address contains invalid characters" );
                bytes[i] = bytes[i] << 4 | cval;
            }
            if ( *p != '\0' && *p++ != ':' )
                throw invalid_argument( "format of ascii hardware address invalid" );
        }
    }
    
    if ( *p )
        throw invalid_argument( "ascii hardware address too long" );
    
    checkValidity();
}

/**
 * Makes a copy of the first <size> bytes of the argument. 
 * Size must be <= the size of hw.
 */
HwAddr::HwAddr(const HwAddr &hw, int size) throw(std::invalid_argument)
{    
    if(size > hw.addrsize)
        throw std::invalid_argument("Size must be <= original size.");
    
    addrsize = size;
    memset( (void*)bytes, 0, 6 );
    strncpy( tag, hw.tag, sizeof( tag ) );

    cpyMem(hw.bytes);
}

void
HwAddr::cpyMem(const unsigned char *buf) throw(std::invalid_argument)
{
    memcpy( (void*)bytes, buf, addrsize );
}

unsigned char
HwAddr::hexval(const char c) throw()
{
    if ( c >= '0' && c <= '9' )
        return c - '0';
    else if ( c >= 'A' && c <= 'F' )
        return c - 'A' + 10;
    else if ( c >= 'a' && c <= 'f' )
        return c - 'a' + 10;
    
    return 0xff;
}

HwAddr &
HwAddr::operator=(const HwAddr &obj) throw(std::invalid_argument)
{
    memcpy( (void*)bytes, (void*)obj.bytes, 6 );
    memcpy( tag, obj.tag, sizeof(tag) );
    addrsize = obj.addrsize;
    
    checkValidity();
    
    return *this;
}


void
HwAddr::checkValidity() const throw(std::invalid_argument)
{
    unsigned char v;
    
    // Check for all zeros (invalid).
    v = 0;
    for ( int i = 0; i < addrsize; ++i )
        v = v | bytes[i];
    
    if ( v == 0 )
        throw invalid_argument( "invalid hardware address");
    
    // Check for all ones (broadcast destination).
    v = 0xff;
    for ( int i = 0; i < addrsize; ++i )
        v = v & bytes[i];
    
    if ( v == 0xff )
        throw invalid_argument( "broadcast hardware address not allowed" );
    
    // Check for least-significant-bit of first byte is 1
    // (multicast destinations)
    //if ( (bytes[0] & 0x01) != 0 )
        //throw invalid_argument( "multicast hardware addresses not allowed" );
}

unsigned char
HwAddr::operator[](const int i) const throw(std::out_of_range)
{
    if ( i > -1 && i < 6 )
        return bytes[i];
    
    throw std::out_of_range( "valid range for HwAddr: 0-5" );
}

size_t
HwAddr::hashkey() const throw()
{
    size_t key = 0;
    static const short shiftvals[6] = { 0, 5, 5, 5, 6, 6 };
    
    for ( int i = 0; i < 6; ++i )
        key = (key << shiftvals[i]) ^ (size_t)bytes[i];
    
    return key;
}

bool HwAddr::operator==(const HwAddr &hw) const throw()
{
    return (memcmp( (void*)bytes, (void*)hw.bytes, addrsize ) == 0 );
}

string
HwAddr::ascii() const throw()
{
    string asc;
    
    char val[3];
    
    for(int i = 0; i < addrsize; i++)
    {
        if(!asc.empty())
            asc += ":";
        snprintf( val, 3, "%02hhX", bytes[i] );
        asc += val;
    }
    
    return asc;
}

int HwAddr::size() const
{
    return addrsize;
}

void HwAddr::setTag(const char* tag)
{
    if(tag)
    {
        memset( this->tag, '\0', sizeof( tag ) );
        strncpy( this->tag, tag, sizeof( this->tag )-1 );
    }
}

const char* HwAddr::getTag() const
{
    return tag;
}

// ---- HwAddrHash Class Definition --------------

HwAddrHash::HwAddrHash(int minBuckets) : m_mutex( "HwAddrHash mutex" ),
        m_hash(minBuckets)
{
}

HwAddrHash::~HwAddrHash()
{
}


void
HwAddrHash::put(const HwAddr &hw, int val/*=1*/)
{
    LOCK();
    const HashType::value_type v = HashType::value_type(hw, val);
    m_hash.erase(hw);
    m_hash.insert( v );
}

bool
HwAddrHash::get(HwAddr &hw, int &val) const
{
    HashType::const_iterator i;
    
    LOCK();
    
    if ( (i = m_hash.find( hw )) != m_hash.end() )
    {
        hw = (*i).first;
        val = (*i).second;
        return true;
    }
    
    return false;
}

bool
HwAddrHash::remove(const HwAddr &hw)
{
    HashType::iterator i;
    bool found = false;
    
    LOCK();
    
    if ( (i = m_hash.find( hw )) != m_hash.end() )
    {
        found = true;
        m_hash.erase( i );
    }
    
    return found;
}

bool
HwAddrHash::exists(const HwAddr &hw) const
{
    LOCK();
    
    return (m_hash.find(hw) != m_hash.end());
}

bool
HwAddrHash::getAll(HwAddrHash::AddrList &alist) const
{
    LOCK();
    
    HashType::const_iterator i = m_hash.begin();
    const HashType::const_iterator iEnd = m_hash.end();
    
    for ( ; i != iEnd; ++i )
        alist.push_back( (*i).first );
    
    return (! alist.empty());
}

void
HwAddrHash::clear()
{
    LOCK();
    
    m_hash.clear();
}

string
HwAddrHash::generateStats() const
{
    string result;
    char x[100];
    
    result.reserve(500);
    
    unsigned int size, nbuckets;
    int max = -1, min = 999999;
    
    // nelems_in_bucket: idx = bucket#, value = elts in that bucket
    vector<int> nelems_in_bucket;
    
    {
        LOCK();
        size = m_hash.size();
        nbuckets = m_hash.bucket_count();
        
        for ( unsigned int i = 0; i < m_hash.bucket_count(); ++i )
        {
            int x = m_hash.elems_in_bucket( i );
            if ( x > max ) max = x;
            if ( x < min ) min = x;
            nelems_in_bucket.push_back( x );
        }
    }
    
    result += "  Number of entries in hash: ";
    result += (sprintf( x, "%d\n", size ), x );
    
    result += "  Number of buckets: ";
    result += (sprintf( x, "%d\n", nbuckets ), x);
    
    result += (sprintf( x, "  Bucket depth range: [%d, %d]\n", min, max ), x );
    
    if ( max > 1000 )
    {
        result += "  max depth > 1000??? BAD!\n";
        return result;
    }
    
    // buckets_of_depth - index = depth; value = n buckets with that depth
    int buckets_of_depth[max+1];
    memset( (void*)buckets_of_depth, 0, (max+1)*sizeof(int) );
    
    for ( unsigned int i = 0; i < nbuckets; ++i )
        buckets_of_depth[nelems_in_bucket[i]]++;
    
    result += "  Quantity of buckets at each depth:\n";
    
    int modeval = -1, mode = -1;
    for ( int i = 0; i <= max; ++i )
    {
        result += (sprintf( x, "    %6d (%6.2f%%) buckets of depth %d\n",
                buckets_of_depth[i],
                (double)buckets_of_depth[i]*100.0/(double)nbuckets,
                i ), x);
        
        if ( buckets_of_depth[i] > modeval )
        {
            modeval = buckets_of_depth[i];
            mode = i;
        }
    }
    
    // Sort the vector of number of items in each bucket so that we can
    // find the median.
    sort( nelems_in_bucket.begin(), nelems_in_bucket.end() );
    result += (sprintf(x, "  Mean: %.2f\n", (double)size/(double)nbuckets), x);
    result += (sprintf(x, "  Median: %d\n", nelems_in_bucket[nbuckets/2] ), x);
    result += (sprintf(x, "  Mode: %d\n", mode ), x);
    
    return result;
}


// ---- Inner Class (functor) Definition --------------

size_t
HwAddrHash::HashFunc::operator()(const HwAddr &hw) const
{
    return hw.hashkey();
}

bool
HwAddrHash::EqualKey::operator()(const HwAddr &hw1, const HwAddr &hw2) const
{
    return (hw1 == hw2);
}

// ---- Unit test definition -------------------

#ifdef UNITTEST
void HwAddr::test() throw()
{
    bool pass = true;
    string macs[6];
    macs[0] = string("02");  // All zeros is a special case and should fail
    macs[1] = string("00:11");
    macs[2] = string("00:11:22");
    macs[3] = string("00:11:22:33");
    macs[4] = string("00:11:22:33:44");
    macs[5] = string("00:11:22:33:44:55");
    
    
    for(int i = 0; i < 6; i ++)
    {
        HwAddr hw(macs[i]);
        
        // Check resultant mac size
        if( hw.size() != i + 1 )
            throw "Improper size";
        
        if(!(hw.ascii() == macs[i]) )
            throw "Not equal after creation";
    }
    
    try
    {
        HwAddr hw("00");
        pass = false;
    }
    catch (invalid_argument e)
    {
        pass = true;
    }
    
    if(!pass)
        throw "Should have failed with all zeros";
}
#endif // UNITTEST
