//***************************************************************************
//
// $Source$
// $Revision: 7388 $
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: Mutex
//     Author: Jim Guggemos
//     Description: Definition for the Mutex and associated class
//
// $Date: 2006-02-06 09:04:13 -0700 (Mon, 06 Feb 2006) $
//
//***************************************************************************

#ifndef __Mutex_H__
#define __Mutex_H__

#include <pthread.h>
#include <semaphore.h>
#include <string>
#include <errno.h>
#include <stdio.h>

using namespace std;

class ConditionVar;


/** Wrapper class for a pthread_mutex.  Provides a "resource acqusition is
    initialization" helper inner class, Mutex::Holder.
*/
class Mutex
{
public:
    /** Different types of pthread mutex:
	- FAST = if owning thread locks more than once, deadlock ensues.
	- SAFE = if owning thread locks more than once, lock() return false.
	- RECURSIVE = owning thread must unlock as many times as it locks.
	
	NOTE: Mutexes REQUIRE that the thread that locks it must be the
	same thread that unlocks it.  It is not checked for FAST mutexes in
	the Linux implementation but that is not portable.  If you want 
	mutex-like behavior but where different threads may lock and
	unlock the mutex, use a Semaphore, defined below.  Otherwise,
	especially if you use RECURSIVE or SAFE mutexes, you will not get
	the behavior you expect. */
    enum MutexType { FAST, SAFE, RECURSIVE };

    /// Construct an unnamed Mutex.
    Mutex(MutexType type=FAST);

    /// Construct a named Mutex.
    Mutex(const string &name, MutexType type=FAST);

    /// Destructor
    virtual ~Mutex();

    /** Lock mutex.  Always returns true unless this is a SAFE mutex.  In the
	case of a SAFE mutex, if the owning thread already has it locked,
	it will return false.  */
    inline bool lock() { return (pthread_mutex_lock( &m_mutex ) != EDEADLK); }

    /// Unlock mutex
    inline void unlock() { pthread_mutex_unlock( &m_mutex ); }

    /// Return name
    inline const string &name() { return m_name; }


    /** "Resource acquisition is initialization" helper class.  Construct
	one of these on the stack and it will lock upon construction and
	unlock upon destruction.  */
    class Holder
    {
    private:
	Mutex &m_mutex;

    public:
	Holder( Mutex &mutex );
	Holder( ConditionVar &condx );
	~Holder();
    };

protected:
    friend class ConditionVar;

    /// The actual mutex
    pthread_mutex_t m_mutex;

private:
    /// Name for debugging purposes
    const string m_name;

    /// Initialize the object
    void init(MutexType type);
};


/** Wrapper around a POSIX condition variable.  A condition variable
    requires the use of a Mutex.  See pthread_cond_init(3) for more details
    on how to use this.  Note that you need to hold the mutex before
    calling signal, broadcast or wait and thus a call to any of those
    may block.  Therefore, these shouldn't be used in signal handlers.
    However, the nice feature is that wait has an optional timeout.

    There are two sets of constructors -- one automatically contructs
    and destroys its own (FAST) Mutex; the other takes a Mutex and does
    NOT destroy it upon the destruction of the ConditionVar. */
class ConditionVar
{
public:
    /// Construct an unnamed ConditionVar with a locally constructed Mutex
    ConditionVar();

    /// Construct a named ConditionVar with a locally constructed Mutex
    ConditionVar(const string &name);

    /// Construct an unnamed ConditionVar with given Mutex
    ConditionVar(Mutex *mutex);

    /// Construct a named ConditionVar with given Mutex
    ConditionVar(const string &name, Mutex *mutex);

    /// Destructor
    virtual ~ConditionVar();

    /// Lock the associated Mutex (see Mutex::lock() for details.)
    inline bool lock() { return m_mutex->lock(); }

    /// Unlock the associated Mutex (see Mutex::unlock() for details.)
    inline void unlock() { m_mutex->unlock(); }

    /// Returns access to the mutex.
    Mutex &mutex() { return *m_mutex; }

    /** Wake EXACTLY one of the threads waiting on the ConditionVar; if no
	threads are waiting, nothing happens.
	NOTE THAT MUTEX MUST BE LOCKED BEFORE CALLING THIS METHOD.  */
    inline void signal() { pthread_cond_signal( &m_cond ); }

    /** Wake ALL of the threads waiting on the ConditionVar; if no threads
	are waiting, nothing happens.
	NOTE THAT MUTEX MUST BE LOCKED BEFORE CALLING THIS METHOD.  */
    inline void broadcast() { pthread_cond_broadcast( &m_cond ); }

    /** Wait (block) on the ConditionVar.  If a timeout is specified, the
	thread will wake up after timeout seconds if no other thread signaled
	the condition.  Returns true if condition was signaled, false if a 
	timeout occurred.
	NOTE THAT MUTEX MUST BE LOCKED BEFORE CALLING THIS METHOD and
	that the Mutex will be unlocked atomically before blocking and
	reacquired before returning.  */
    bool wait(int timeout=0);

protected:
    /// The Mutex required by condition variables.
    Mutex *m_mutex;

    /// True if we own the Mutex - this it must be deleted upon destruction.
    bool m_ourMutex;

    /// Name for debugging purposes
    const string m_name;

    /// The actual condition variable
    pthread_cond_t m_cond;
};


/** Wrapper for a POSIX semaphore.  Note that on most architectures,
    these are asynch-signal safe.  They are, infact, the only POSIX thread
    synchronization method that is asynch-signal safe. */
class Semaphore
{
public:
    /** Allocate a Semaphore with the given initial_value.  Use
	1 for initial_value if you want mutex-like behavior. */
    Semaphore(int initial_value=0);

    /// Allocate a named semaphore
    Semaphore(const string &name, int initial_value=0);

    /// Destructor
    ~Semaphore();

    /** Blocks thread until semaphore has non-zero count; then atomically
	decrements the counter. Returns false if some error other than
	EINTR occurred (will re-wait on EINTR).  If false is returned,
	the "wait" operation did not succeed. */
    bool wait();

    /** Non-blocking version of wait.  Returns true if the count was
	decremented successfully; false if the count was already 0. */
    bool wait_non_blocking() { return (sem_trywait( &m_semaphore ) == 0); }

    /** Atomically increments the count, resulting in a thread waking
	if one was waiting on the semaphore.  Guaranteed not to block. */
    inline void post() { sem_post( &m_semaphore ); }

    /// Return the semaphore count; be wary in using this except for debugging.
    int count() { int x; sem_getvalue( &m_semaphore, &x ); return x; }

    /// Return name
    inline const string &name() { return m_name; }

protected:
    void init(int initial_value);

    sem_t m_semaphore;

private:
    const string m_name;

};


#endif // __Mutex_H__
