//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: PacketQueue
//     Author: Jim Guggemos
//     Description: Class to interface with libnetfilter_queue
//
// $Date$
//
//***************************************************************************

#include "PacketQueue.h"
#include "SAPQ.h"
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>


/// ---------------------------------------------------------------------------
/// --- BEGIN PacketDispatcher ------------------------------------------------
/// ---------------------------------------------------------------------------

// If you have to synchronize on both, sync on NFQ first, then QUEUE_SET.

#define SYNCHRONIZE_NFQ() Mutex::Holder _nfqLock(this->m_nfqMutex)
#define SYNCHRONIZE_QUEUE_SET() Mutex::Holder \
    _queueSetLock(this->m_queueSetMutex)

PacketDispatcher::PacketDispatcher(int maxPacketSize) throw(PacketQueueError) :
    m_queueSet(), m_queueSetMutex( "PacketDispatcher::m_queueSetMutex" ),
    m_nfqMutex("PacketDispatcher::m_nfqMutex"),
    m_closeCondition("PacketDispatcher::m_closeCondition"),
    m_stop(false), m_closed(false), m_nfqHandle(0), m_bufsiz(maxPacketSize)
{
    if ( ! (m_nfqHandle = nfq_open()) )
	throw PacketQueueError("PacketDispatcher",
                               "Could not create NFQ handle");

    try
    {
        if ( nfq_unbind_pf( m_nfqHandle, AF_INET ) < 0 )
            throw PacketQueueError("PacketDispatcher",
                   "Could not unbind existing NF_QUEUE handler for AF_INET");

        if ( nfq_bind_pf( m_nfqHandle, AF_INET ) < 0 )
            throw PacketQueueError("PacketDispatcher",
             "Could not bind nfnetlink_queue as NF_QUEUE handler for AF_INET");

        m_buf = new unsigned char[m_bufsiz];

        if ( (m_nfqFD = nfnl_fd( nfq_nfnlh( m_nfqHandle ) )) < 0 )
            throw PacketQueueError("PacketDispatcher",
                                   "Could not NFQ file descriptor");
    }
    catch ( PacketQueueError &ex )
    {
        // Catch any error thrown here and clean up system resources;
        // then re-throw the bad boy.
        close();
        throw ex;
    }
}

PacketDispatcher::~PacketDispatcher() throw()
{
    if ( ! isClosed() )
        close();
}

void
PacketDispatcher::addQueue(PacketQueue *queue)
    throw(PacketQueueError)
{
    {
        SYNCHRONIZE_NFQ();
        // This can throw a PacketQueueError
        queue->createQueueHandle( this, m_bufsiz );
    }

    {
        SYNCHRONIZE_QUEUE_SET();
        m_queueSet.insert( queue );
    }
}

void
PacketDispatcher::removeQueue(PacketQueue *queue) throw()
{
    {
        SYNCHRONIZE_QUEUE_SET();
        m_queueSet.erase( queue );
    }
}

void
PacketDispatcher::run() throw()
{
    int rv;

    // The main run loop will timeout on select
    while ( ! m_stop )
    {
        fd_set readfds;
        struct timeval tv;

        FD_ZERO( &readfds );
        FD_SET( m_nfqFD, &readfds );

        tv.tv_sec = 2;
        tv.tv_usec = 0;

        int sel_ret = select( m_nfqFD+1, &readfds, NULL, NULL, &tv );
        if ( sel_ret < 0 )
        {
            char errbuf[512];
            logMessage("PacketDispatcher::run : select(): %s",
                       errnoString(errno, errbuf));
            m_stop = true;
        }

        else if ( sel_ret > 0 && FD_ISSET( m_nfqFD, &readfds ) )
        {
            if ((rv = recv( m_nfqFD, (char*)m_buf, m_bufsiz, 0 )) >= 0)
            {
                SYNCHRONIZE_NFQ();
                nfq_handle_packet( m_nfqHandle, (char*)m_buf, rv );
            }
            else
            {
                char errbuf[512];
                logMessage("PacketDispatcher::run : recv(): %s", errnoString(errno, errbuf));

                if(errno != ENOBUFS) // Warning that kernel is out of buffer space, keep going
                    m_stop = true;
            }
        }

        {
            SYNCHRONIZE_QUEUE_SET();
            if ( m_queueSet.empty() )
                m_stop = true;
        }
    }

    close();
}

void
PacketDispatcher::stop() throw()
{
    m_stop = true;

    m_closeCondition.lock();

    while ( ! m_closed )
        m_closeCondition.wait();

    m_closeCondition.unlock();
}

void
PacketDispatcher::close() throw()
{
    {
        SYNCHRONIZE_QUEUE_SET();

        for ( set<PacketQueue *>::iterator pqI = m_queueSet.begin();
              pqI != m_queueSet.end(); ++pqI )
            delete (*pqI);

        m_queueSet.clear();
    }

    if ( m_nfqHandle )
    {
        nfq_close( m_nfqHandle );
        m_nfqHandle = 0;
    }

    if ( m_buf )
    {
	delete m_buf;
	m_buf = 0;
    }

    m_closeCondition.lock();
    m_closed = true;
    m_closeCondition.signal();
    m_closeCondition.unlock();
}

bool
PacketDispatcher::isClosed() throw()
{
    bool closed;
    m_closeCondition.lock();
    closed = m_closed;
    m_closeCondition.unlock();

    return closed;
}

// To use this handle, caller must hold NFQ lock
struct nfq_handle *
PacketDispatcher::handle() throw()
{
    return m_nfqHandle;
}


/// ---------------------------------------------------------------------------
/// --- BEGIN PacketQueue -----------------------------------------------------
/// ---------------------------------------------------------------------------

PacketQueue::PacketQueue(int queueNum, PacketQueueListener *listener)
    throw(PacketQueueError) :
    m_listener(listener), m_queueNum(queueNum), m_queueHandle(0),
    m_dispatcher(0)
{
    if ( m_listener == 0 )
        throw PacketQueueError("PacketQueue",
                               "Listener parameter cannot be null");

    if ( m_queueNum < 0 || m_queueNum > 65535 )
        throw PacketQueueError("PacketQueue",
                               "queueNum parameter out of range");
}

PacketQueue::~PacketQueue() throw()
{
    if ( m_queueHandle )
    {
        nfq_destroy_queue( m_queueHandle );
        m_queueHandle = 0;
    }
}

// MUST be called with NFQ lock held by dispatcher
void
PacketQueue::createQueueHandle(PacketDispatcher *dispatcher, int bufsiz)
    throw(PacketQueueError)
{
    m_dispatcher = dispatcher;

    if ( ! (m_queueHandle = nfq_create_queue(dispatcher->handle(), m_queueNum,
                                             callback_stub, (void*)this)) )
	throw PacketQueueError("PacketQueue", "Could not create NFQ queue");

    if ( nfq_set_mode( m_queueHandle, NFQNL_COPY_PACKET, 1024 ) < 0 )
	throw PacketQueueError("PacketQueue", "Could not set NFQ mode");
}

bool
PacketQueue::callListener(unsigned char** payload, int *len) throw()
{
    // NOTE that is is called with the dispatcher's NFQ lock held.
    bool accept = false;
    try
    {
        accept = m_listener->onPacket( payload, len );
    }
    catch ( ... )
    {
        // If we caught an exception, remove this queue from the
        // dispatcher's queue list and delete myself!
        if ( m_dispatcher )
            m_dispatcher->removeQueue( this );
        delete this;
    }

    return accept;
}

int
PacketQueue::callback_stub(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg,
                           struct nfq_data *nfa, void *data)
{
    PacketQueue *this_pq = reinterpret_cast<PacketQueue *>(data);
    bool accept = false;
    unsigned char *payload = 0;
    int len = nfq_get_payload( nfa, &payload );
    
    if (this_pq == 0)
    {
        logMessage("PacketQueue::callback_stub called without PacketQueue");
    }

    else
    {
        accept = this_pq->callListener( &payload, &len );
    }

    struct nfqnl_msg_packet_hdr *ph;
    int id = -1;

    if ( (ph = nfq_get_msg_packet_hdr( nfa )) )
        id = ntohl( ph->packet_id );

    int verdict = (accept ? NF_ACCEPT : NF_DROP);
    int rval = (id > -1 ? nfq_set_verdict( qh, id, verdict, len, payload ) : -1);

    /*
    fprintf(stderr, "_stub: accept: %d verdict: %d id: %d rval: %d\n",
            accept, verdict, id, rval);
    */

    return rval;
}


/// ---------------------------------------------------------------------------
/// --- BEGIN PacketQueueError ------------------------------------------------
/// ---------------------------------------------------------------------------

PacketQueueError::PacketQueueError(const std::string &method,
                                   const std::string &message)
{
    m_text = "error occured in method " + method + ": " + message;
}
