//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Class: SocketServer
//     Author: Jim Guggemos
//     Description: Class that monitors listening socket and creates
//                  Client objects (and thread) when a client connects
//                  via TCP socket.
//
// $Date$
//
//***************************************************************************

#include "SocketServer.h"
#include "Socket.h"
#include "SAPQ.h"
#include "Client.h"

SocketServer::SocketServer(HashManager &hashmgr, int port) :
    m_port(port), m_threadId(0), m_stop(false), m_hashmgr(hashmgr), m_listenSocket(0)
{
}

SocketServer::~SocketServer()
{
    stop();
}

bool
SocketServer::start()
{
    ServiceMonitor::instance()->setSocketServer( true );

    pthread_attr_t attr;
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
    if ( pthread_create( &m_threadId, &attr, threadRun, (void*)this ) != 0 )
    {
        ServiceMonitor::instance()->setSocketServer( false );
        return false;
    }
    pthread_attr_destroy( &attr );  // No effect in LINUX

    return true;
}

void
SocketServer::stop()
{
    m_stop = true;

    logMessage( "SocketServer stopped" );

    ServiceMonitor::instance()->setSocketServer( false );

    if ( m_threadId > 0 )
    {
        pthread_t die = m_threadId;
        m_threadId = 0;

        pthread_cancel( die );
        // We may have killed ourselves here, by the way.  Don't do anything
        // important after this call!
    }
}

void
SocketServer::run()
{
    logMessage( "SocketServer started" );
    try
    {
        m_listenSocket = new Socket();
        m_listenSocket->bind( Socket::ADDR_LOCALHOST, m_port );
        m_listenSocket->listen( 10 );
        logDebug( "SocketServer socket open, bound to port %d", m_port );
    }
    catch (Socket::Exception &e)
    {
        logMessage( "SocketServer failed to initialize: %s",
                    e.getMsg().c_str() );
        stop();
        return;
    }
    
    // Will delete socket when this scope exists, including when thread
    // is killed.  This will also close all other sockets that are currently
    // open.
    class sshelper
    {
    private:
        Socket *m_sock;
        
    public:
        sshelper(Socket *s) : m_sock(s) {}
        ~sshelper() { delete m_sock; Socket::closeAllImmediate(); }
    } _sock( m_listenSocket );


    while ( ! m_stop )
    {
        Socket *newSock;

        try
        {
            newSock = m_listenSocket->accept();
        }
        catch (Socket::ClosedException)
        {
            logMessage( "SocketServer listening socket was closed." );
            stop();
            return;
        }
        catch (Socket::Exception &e)
        {
            logMessage( "SocketServer exception on listening socket: %s",
                        e.getMsg().c_str() );
            stop();
            return;
        }

        logDebug( "SocketServer has accepted a new connection" );
        // Client will kill itself once the socket closes.
        Client *newClient = new Client( m_hashmgr, newSock );
        if ( ! newClient->start() )
        {
            logMessage( "SocketServer failed to start thread for client" );
            delete newClient;
            delete newSock;
        }
    }
}



void *
SocketServer::threadRun(void *vthis)
{
    SocketServer *rthis = reinterpret_cast<SocketServer *>(vthis);

    rthis->run();
    return NULL;
}
