//***************************************************************************
//
// $Source$
// $Revision$
//
//
//   Latis Networks, Inc. (d/b/a StillSecure)
//     www.stillsecure.com
//     www.stillsecure.org
//
//
//   Copyright (C) 2005-2006 Latis Networks, Inc. (d/b/a StillSecure)
//
//    This file is part of SAPQ.
//
//    SAPQ is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    SAPQ is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAPQ; if not, write to the Free Software
//    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
//
//
//     Project: Safe Access Packet Queuing Daemon
//     Module: SAPQ
//     Author: Jim Guggemos
//     Description: SAPQ main() and associated private utility functions.
//
// $Date$
//
//***************************************************************************

#include "PacketQueue.h"
#include "HwAddrHash.h"
#include "HashManager.h"
#include "DHCPWatcher.h"
#include "SocketServer.h"
#include "SAPQ.h"
#include "version.h"

#include <iostream>
#include <stdarg.h>
#include <syslog.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>

#define SYSLOG_LEVEL LOG_ERR
#define SYSLOG_FACILITY LOG_LOCAL3

#define DEFAULT_PORT 5341
#define DEFAULT_QUEUE_NUM 0

ServiceMonitor* ServiceMonitor::m_instance = new ServiceMonitor();

namespace sapq
{
    bool debug = false;
    bool daemon = false;
    bool reverse = false;
    int port = DEFAULT_PORT;
    int queueNum = DEFAULT_QUEUE_NUM;
    string pidFile;
    string versionBanner;
    string shortCopyrightBanner;
    string copyrightBanner;
    int minBuckets = 1571;
    
    void makePidFile();
    void removePidFile();
    void initSigHandlers();
    void signalHandler(int sig);
    void usage();
    void parseArgs(int argc, char *argv[]);
    void log(const char *format, va_list ap);
    string logFormat(const char *format, va_list ap);
}

#ifdef UNITTEST
int
main(int argc, char **argv)
{
    try
    {
        HwAddr::test();
        HashManager::test();
    }
    catch(const std::logic_error &err)
    {
        printf("%s", err.what());
        return 1;
    }
    catch(...)
    {
        printf("Failed with unknown error\n");
        return 1;
    }
    printf("Pass\n");
    return 0;
}

#else // Not UNITTEST

int
main(int argc, char *argv[])
{
    sapq::parseArgs( argc, argv );
    
    if ( sapq::daemon )
    {
        if ( fork() > 0 )
            exit( 0 );
        
        setsid();
        close( 0 );
        close( 1 );
        close( 2 );
    }
    
    sapq::initSigHandlers();
    
    logMessage( "%s", sapq::versionBanner.c_str() );
    logDebug( "sapq started in %s mode%s",
            (sapq::daemon) ? "daemon" : "foreground",
            (sapq::debug) ? " with debugging output enabled" : "");
    
    HashManager hashmgr( sapq::minBuckets );
    
    DHCPWatcher watcher( hashmgr, sapq::queueNum, sapq::reverse );
    watcher.start();
  
    SocketServer sockServer( hashmgr, sapq::port );
    sockServer.start();
    
    sapq::makePidFile();
    
    logMessage( "sapq ready" );
    ServiceMonitor::instance()->blockUntilNotRunning();
    
    logMessage( "Some service has stopped. Exiting..." );
    
    sockServer.stop();
    watcher.stop();
    
    sleep( 1 );
    
    sapq::removePidFile();
    exit( 0 );
};
#endif // UNITTEST

void
sapq::initSigHandlers()
{
    signal( SIGINT, sapq::signalHandler );
    signal( SIGQUIT, sapq::signalHandler );
    signal( SIGTERM, sapq::signalHandler );
}

void
sapq::signalHandler(int sig)
{
    logMessage( "Received a signal (%s).. shutting down.",
            ((sig == SIGINT) ? "SIGINT" :
                ((sig == SIGQUIT) ? "SIGQUIT" :
                    ((sig == SIGTERM) ? "SIGTERM" : "???"))) );
    
    ServiceMonitor::instance()->interrupt();
}

void
sapq::makePidFile()
{
    if ( ! sapq::pidFile.empty() )
    {
        int fd;
        struct stat st;
        
        if ( stat( sapq::pidFile.c_str(), &st ) == 0 )
        {
            if ( S_ISDIR( st.st_mode ) )
                sapq::pidFile += "/sapq.pid";
        }
        else if ( errno != ENOENT )
        {
            char err[512];
            logMessage( "path given to --pid arg (%s) is bad: %s",
                    sapq::pidFile.c_str(), errnoString(errno, err) );
            
            sapq::pidFile.clear();
            return;
        }
        
        if ( (fd=open( pidFile.c_str(), O_WRONLY|O_CREAT|O_TRUNC, 0644 )) < 0 )
        {
            char err[512];
            logMessage( "Could not write pid file '%s': %s",
                    pidFile.c_str(), errnoString(errno, err) );
            sapq::pidFile.clear();
        }
        
        else
        {
            char pids[512];
            sprintf( pids, "%d\n", getpid() );
            write( fd, pids, strlen(pids) );
            close( fd );
        }
    }
}

void
sapq::removePidFile()
{
    if ( ! sapq::pidFile.empty() )
    {
        unlink( sapq::pidFile.c_str() );
        sapq::pidFile.clear();
    }
}

void
sapq::usage()
{
    cerr << sapq::versionBanner << endl;
    
    cerr << "Usage: sapq [--debug] [--daemon] [--pid=<fname>] [--port=<port>]"
            << endl
            << "            [--queue-num=<q>] [--buckets=<b>]" << endl
            << endl
            << " --debug         Logs debugging information" << endl
            << " --daemon        Starts as a daemon, logs to syslog (without this"
            << endl
            << "                   option, logging is sent to stderr)" << endl
            << " --pid=<f>       Create a pid file with the filename <f>" << endl
            << " --port=<p>      Use specified port; default: " << DEFAULT_PORT
            << endl
            << " --reverse       Deny access to allowed endpoints (used when running"
            << endl
            << "                 sapq on the INPUT chain to prevent allowed endpoints"
            << endl
            << "                 from reaching the onboard DHCP server)."
            << endl
            << " --queue-num=<q> Use specified NFQUEUE queue number; default: "
            << DEFAULT_QUEUE_NUM << endl
            << " --buckets=<b>   Use specified minimum number of initial buckets"
            << endl
            << "                 in the hash table (default: " << sapq::minBuckets
            << ")" << endl
            << " --copyright     To show copyright information" << endl;
}

void
sapq::parseArgs(int argc, char *argv[])
{
    {
        char buf[1024];
        snprintf( buf, 1024, "sapq: v%s.%s, %s",
                MAJOR_VERSION, MINOR_VERSION, BUILD_DATE );
        sapq::versionBanner = buf;
        
        sapq::shortCopyrightBanner =
                "Copyright (C) 2005-2006 Latis Networks, Inc. "
                "(d/b/a StillSecure).\n"
                "sapq comes with ABSOLUTELY NO WARRANTY.  This is free software, "
                "and you\n"
                "are welcome to redistribute it under certain conditions; type\n"
                "'sapq --copyright' for details.";
        
        sapq::copyrightBanner =
                "Copyright (C) 2005-2006 Latis Networks, Inc. "
                "(d/b/a StillSecure).\n"
                "\n"
                "This program is free software; you can redistribute it and/or\n"
                "modify it under the terms of the GNU General Public License\n"
                "as published by the Free Software Foundation; either version 2\n"
                "of the License, or (at your option) any later version.\n"
                "\n"
                "This program is distributed in the hope that it will be useful,\n"
                "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
                "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
                "GNU General Public License for more details.\n"
                "\n"
                "You should have received a copy of the GNU General Public License\n"
                "along with this program; if not, write to the Free Software\n"
                "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."
                "\n\n"
                "See http://www.gnu.org/licenses/ for more information on the GPL.";
    }
    
    for ( char **argp = &argv[1]; *argp; ++argp )
    {
        string arg( *argp );
        const int len = arg.length();
        
        if ( arg == "--debug" )
            sapq::debug = true;
        
        else if ( arg == "--daemon" )
        {
            sapq::daemon = true;
            openlog( "sapq", LOG_PID, SYSLOG_FACILITY );
        }
        
        else if ( arg == "--copyright" )
        {
            fprintf( stderr, "%s\n%s\n", sapq::versionBanner.c_str(),
                    sapq::copyrightBanner.c_str() );
            exit( 0 );
        }
        
        else if ( len > 6 && arg.substr( 0, 6 ) == "--pid=" )
            sapq::pidFile = arg.substr( 6 );
        
        else if ( len > 12 && arg.substr( 0, 12 ) == "--queue-num=" )
        {
            sapq::queueNum = atoi( arg.substr(12).c_str() );
            if ( sapq::queueNum < 0 || sapq::queueNum > 65535 )
            {
                cerr << "Invalid queue number given (" << sapq::queueNum << ")."
                        << endl;
                usage();
                exit( 1 );
            }
        }
        
        else if ( len > 10 && arg.substr( 0, 10 ) == "--buckets=" )
            sapq::minBuckets = atoi( arg.substr( 10 ).c_str() );
        
        else if ( len > 7 && arg.substr( 0, 7 ) == "--port=" )
        {
            sapq::port = atoi( arg.substr(7).c_str() );
            if ( sapq::port < 1 || sapq::port > 65535 )
            {
                cerr << "Invalid port number given (" << sapq::port << ")."
                        << endl;
                usage();
                exit( 1 );
            }
        }
        else if ( arg == "--reverse" )
            sapq::reverse = true;
        
        else
        {
            usage();
            exit( 1 );
        }
    }
    
    if ( getuid() != 0 )
    {
        sapq::daemon = false;  // log to stderr
        logMessage( "sapq must be run as root" );
        exit( 1 );
    }
    
}


void
sapq::log(const char *format, va_list ap)
{
    if ( sapq::daemon )
        vsyslog( SYSLOG_LEVEL, format, ap );
    else
    {
        string msg = sapq::logFormat( format, ap );
        fprintf( stderr, "%s\n", msg.c_str() );
    }
}

string
sapq::logFormat(const char *format, va_list ap)
{
    char date[64];
    char fmt[128];
    char buf[4096];
    
    struct tm lt;
    time_t now = time(NULL);
    
    strftime( date, 64, "%b %d %T", localtime_r( &now, &lt ) );
    sprintf( fmt, "%s sapq[%d]: %s", date, getpid(), format );
    vsnprintf( buf, 4096, fmt, ap );
    buf[4095] = 0;
    
    return string( buf );
}

void
logMessage(const char *format, ...)
{
    va_list ap;
    va_start( ap, format );
    
    sapq::log( format, ap );
    
    va_end( ap );
}

void
logDebug(const char *format, ...)
{
    va_list ap;
    va_start( ap, format );
    
    if ( sapq::debug )
        sapq::log( format, ap );
    
    va_end( ap );
}

char *
errnoString(int errnoValue, char *buf)
{
    // Use re-entrant strerror function, handling unknown value.
    // This "GNU" form of strerror_r may use the passed buffer
    // and it may not; either way it returns a pointer to
    // the string and is re-entrant (hopefully!).
    char *rbuf = strerror_r( errnoValue, buf, sizeof(buf)-1 );
    if ( rbuf == NULL )
        rbuf = "???";
    
    return rbuf;
}

bool
debugMode()
{
    return sapq::debug;
}

void
setDebugMode(bool mode)
{
    sapq::debug = mode;
}


ServiceMonitor::ServiceMonitor() :
                        m_condx( "ServiceMonitor condx" ), m_dhcpWatcherRunning(false),
        m_socketServerRunning(false), m_interrupted(false)
{
}

void
ServiceMonitor::setDHCPWatcher(bool running)
{
    m_condx.lock();
    m_dhcpWatcherRunning = running;
    m_condx.signal();
    m_condx.unlock();
}


void
ServiceMonitor::setSocketServer(bool running)
{
    m_condx.lock();
    m_socketServerRunning = running;
    m_condx.signal();
    m_condx.unlock();
}

void
ServiceMonitor::interrupt()
{
    m_condx.lock();
    m_interrupted = true;
    m_condx.signal();
    m_condx.unlock();
}

void
ServiceMonitor::blockUntilNotRunning()
{
    m_condx.lock();
    
    while ( m_dhcpWatcherRunning && m_socketServerRunning && !m_interrupted)
        m_condx.wait();
    
    m_condx.unlock();
}
